#include<iostream>
#include "DistSim.h"
using namespace std;
		
#define ENDED 2
#define EXECUTED 3

DistSim :: DistSim(){
	Go = NULL;
	Srt = NULL;
	GoSize = 0;
	Frequence = NULL;
	
}

DistSim :: ~DistSim(){
}

void DistSim :: setDelta(float delta){
	dt = delta;
}


int DistSim :: xorChoice(int max){
	
	double value, x;
	int i;
	x = ((double)rand()/RAND_MAX);
	for(i = 1; i <= max; i++){
		if(x <= (double)i/max)
			return i-1;
	}
}


void DistSim :: orChoice(int max, int tab[]){
	
	double value, x;
	int i, flag_set;
	flag_set = FALSE; // we use this flag to set at least one option to true.
	for(i = 0; i < max; i++){
		x = ((double)rand()/RAND_MAX);
		if(x >= 0.5){
			tab[i] = TRUE;
			flag_set = TRUE;
		}
		else
			tab[i] = FALSE;
	}
	if(flag_set == FALSE){
		x = ((double)rand()/RAND_MAX)*(max-1);	
		i = (int)x;
		tab[i] = TRUE;
	}
	
}

int DistSim :: getNodeNumber(){
	int i;
	int nodeNumber;
	nodeNumber = 0;
	for(i = 0; i < GoSize; i++){
		if(StateExecutable[i] == TRUE)
			nodeNumber++;
	}
	return nodeNumber;
}


void DistSim ::init(){

	int i;
	
	Srt = new float[GoSize];
	StateExecutable = new int[GoSize];
	Transit = new float[GoSize];
	Frequence = new int[GoSize];
	Nloop = new int[GoSize];
	NPred = new int[GoSize];
	for(i = 0; i < GoSize; i++){
		Srt[i] = 0;
		StateExecutable[i] = FALSE;
		Frequence[i] = 0;
		Transit[i] = 0;
		Nloop[i] = 0;
		NPred[i] = 0;
	}
}


void  DistSim :: setExecutableTask(){
	int i;

	for(i = 0; i < GoSize; i++){
		Srt[i] = 0;
		StateExecutable[i] = FALSE;
		Transit[i] = 0;
		Nloop[i] = 0;
		NPred[i] = Go[i].predNumber;
//		cout<<"pred number "<<NPred[i]<<" \n";
		if(Go[i].task->getType() == OPERATION)
			Srt[i] = Go[i].task->getSrt();
	}

	for(i = 0; i < GoSize; i++){	
		if(Go[i].task->getType() == XOR_JOIN)
			NPred[i] = 1;
		if(NPred[i] == 0){
			StateExecutable[i] = TRUE;
			Frequence[i]++;
		}
 /*
			 Tasks without predecessors can be started.
*/
	}
	
	for(i = 0; i < GoSize; i++){
		if(Go[i].task->getType() != OPERATION && StateExecutable[i] == TRUE)
			activateSuccessor(Go[i].task->getId(), Go[i].task->getType());			
	}

}

void DistSim :: setState(int i){
	
	NPred[i]--;
	if(NPred[i] == 0){
		StateExecutable[i] = TRUE; // we activate the successor
		Frequence[i] = Frequence[i]++;
		Nloop[i] = Go[i].nloop;		

	}	
}

void DistSim :: activateSuccessor(int id, int type){
	int i, j, k, flag_OR;
	int x;
	int *tab;
	
	j = getPos(id);
	if(Go[j].succNumber <= 0 && Nloop[j] <= 0)
		return;
	switch(type){
		case XOR_SPLIT:
			x = xorChoice(Go[j].succNumber);// We select one sucessor;
			i = getPos(Go[j].succs[x]->getId());
			setState(i);
			StateExecutable[j] = EXECUTED;
			break;	
		case OR_SPLIT:
			tab = new int[Go[j].succNumber];
			orChoice(Go[j].succNumber, tab);
/*			We update the number of predecessors of the OR join connector
			depending on the number of activated paths.
*/
			k = 0; flag_OR = FALSE;

			while(k < GoSize && !flag_OR){
				if (Go[k].orSplit == Go[j].task){
					flag_OR = TRUE;
				}
				else
					k++;
			}
			if(!flag_OR){
				cout << "error: Missing OR join connector \n";
				return;
			}else{
				x = 0;
				for(i = 0; i < Go[j].succNumber; i++)
					x += tab[i];
				NPred[k] = x;
				for(x = 0; x < Go[j].succNumber; x++){
					if(tab[x] == TRUE){
						i = getPos(Go[j].succs[x]->getId());
						setState(i);				
					}
				}
				StateExecutable[j] = EXECUTED;	
			}
			break;
		case XOR_JOIN:
//			cout << "nloop = " << Nloop[j] << " \n";
			if(Nloop[j] > 0){
				Nloop[j]--;
				i = getPos(Go[j].predLoop->getId());
				StateExecutable[i] = TRUE;
				Srt[i] = Go[i].task->getSrt();
				Frequence[i]++;
				StateExecutable[j] = FALSE;
			}else{
				i = getPos(Go[j].succs[0]->getId());
				setState(i);
				StateExecutable[j] = EXECUTED;
			}
			break;
		default : // Care, we must take into account message costs !
			for(x = 0; x < Go[j].succNumber; x++){
				i = getPos(Go[j].succs[x]->getId());
				setState(i);
				if(Transit[i] == 0 && Go[j].srtTransit != NULL)
					Transit[i] = Go[j].srtTransit[x];
			} 
			StateExecutable[j] = EXECUTED;	
			break;
	}
}


void DistSim :: setGraphSize(int size){
	GoSize = size;
}

void  DistSim :: updateQoSTask(){
// reduce the SRT of the operation in Executable of dt
	int i;
	for(i = 0; i < GoSize; i++){
		
		if(Transit[i] > 0){
			Transit[i] = Transit[i]-dt;
		}
		
		if(StateExecutable[i] == TRUE){
			if(Transit[i] <= 0){
				Srt[i] = Srt[i]-dt;
				if(Srt[i] <= 0)
					StateExecutable[i] = ENDED;				
			}	

		}
	}
	
}


int DistSim :: getPos(int id){
	int i, trouve;
	trouve = FALSE;
	i = 0;
	while(!trouve && i < GoSize){
		if(Go[i].task->getId() == id)
			trouve = TRUE;
		else
			i++;
	}
	return i;
}




void  DistSim :: updateExecutableTask(){
	int i;
	int *Visited;
// One activates the operations successors
	for(i = 0; i < GoSize; i++){

		if(Go[i].task->getType() == OPERATION){
			if(StateExecutable[i] == ENDED)
				activateSuccessor(Go[i].task->getId(), Go[i].task->getType());
		}
	}

// One activates the successor of all connectors.

/* E.g code
	for(i = 0; i < GoSize; i++){
		if((StateExecutable[i] == TRUE) && (Go[i].task->getType() != OPERATION))
			activateSuccessor(Go[i].task->getId(), Go[i].task->getType());
	}

	 We use Visited since the activation of a connector can activate another one. We will 
	repeat the activation until not having a connector activated. */

	Visited = new int[GoSize];
	for(i = 0; i < GoSize; i++)
		Visited[i] = FALSE;
	i = 0; 
	while(i < GoSize){
		if((StateExecutable[i] == TRUE) && (Go[i].task->getType() != OPERATION) && (Visited[i] == FALSE)){
			activateSuccessor(Go[i].task->getId(), Go[i].task->getType());
			Visited[i] = TRUE;
			i = 0;
		}else
			i++;
	}
	delete[] Visited;

}

void DistSim :: setRequestNumber(int n){
	requestNumber = n;
}

void DistSim :: run(){

	double t;
	int nodeNumber;
	int n, i;

	init();
	n = requestNumber;
	t = 0.0;

	while(n > 0){
		setExecutableTask();
		nodeNumber = getNodeNumber();
		while (nodeNumber > 0){
			t = t+dt;
/*			cout<<" number of nodes = "<<nodeNumber<<" \n";
			cout<< "Active tasks \n";
			for(i = 0; i < GoSize; i++){
				if(StateExecutable[i] == TRUE)
					cout << Go[i].task->getId() << " ";
			}
			cout << " \n";
*/
			updateQoSTask();
			updateExecutableTask();
			nodeNumber = getNodeNumber();
		}
		n--;
	}
	meanTime =  t/requestNumber;
}

double DistSim :: getMeanTime(){
	return meanTime;
}

int DistSim ::  getFrequence(int node){
	return Frequence[node];
}

double DistSim ::  getProba(int node){
	return (double)Frequence[node]/requestNumber;
}



void DistSim :: setOperationNode(OperationNode *ptr){
	Go = ptr;
}

void DistSim ::  deduceOrProba(int f1, int f2, float *p1, float *p2, float *p12){
	float x;
	x = (float)(f1+f2-requestNumber);
	*p1 = (f1-x)/requestNumber;
	*p2 = (f2-x)/requestNumber;
	*p12 = x/requestNumber;
}

